import { List } from 'immutable';
import { Dictionary } from 'typescript-collections';
import { forEach } from 'typescript-collections/dist/lib/arrays';
// todo: get lodash/fp import to work
// import * as _ from 'lodash';

// todo: squash bug with "always 0n" indexes
//   in append_all_toplevel_autoind or TNode_with_children_autoind?
//   PortfolioComponent.test_tree shows the wrong behavior
/**
 * I want a (non-binary) Tree for use e.g. in TreeViewer + TreeEditor
 * Let's start with an immutable API, bc that is an important pillar of computational / fp goodness
 *
 * a Tree is just a Node
 *
 * Nodes are/have ...
 *    an property key: K ,
 *    an property data: D ,
 *    asympt-runtime-reducing meta info, like
 *      max depth of child trees
 *      own depth
 *      #of subnodes
 *      ... something for specific orderings?
 *
 * ? what data struct shall hold the children?
 * rn: Hashmap with... roughly constant access via key?
 * idk inhowfar old structures are reused in function returns rn
 *
 * stretch: let it be cut at depth x
 *          either removing nodes below, or squashing them (in _some_ order) into depth x
 */

export type HashMap<K, D> = Dictionary<K, D>;

export type TNodeOrLeaf<K, A> =
  | { kind: "leaf" }
  | TNode<K, A>

export type MbyTree<K, A> = TNodeOrLeaf<K, A>

export type Path<K> = K[] // ? = List<K>

export const TNode_with_children =<K, D>(key: K) => (data: D) => (child_array: TNode<K, D>[]): TNode<K, D> => {
  let res = new TNode<K, D>(key, data);
  child_array.forEach((child: TNode<K, D>) => res.children.setValue(child.id, child));
  Object.freeze(res);
  return res;
}

export class TNode<ID, A> {
  kind: "non_leaf" = "non_leaf";
  id: ID;
  data: A;
  size: bigint = 1n;
  children: HashMap<ID, TNode<ID, A>> = new Dictionary<ID, TNode<ID, A>>();
  max_depth: number = 0;

  constructor(key: ID, data: A) {
    this.id = key
    this.data = data;
  }

  // EXDMPLE / TEST OBJECTS
  static test_tree_1_3: Tree<bigint, string>
    = TNode_with_children<bigint, string> (0n) ("root of 1-3 tree") (
      [new TNode(0n, "child node #1"),
      new TNode(1n, "the label appearing in a path"),
      new TNode(2n, "or a whole task in Todoist"),
      ])

  static test_tree_1_4: Tree<bigint, string> = append_toplevel_autoind(TNode.test_tree_1_3, new TNode(2n, "2nd root?"))
}

export type Tree<ID, D> = TNode<ID, D>
export type AutoIndTree<D> = TNode<bigint, D>
// TODO ^
export type AutoIndNode<D> = AutoIndTree<D>

export const TNode_with_children_autoind = <D>(data: D) => (child_array: AutoIndNode<D>[]): AutoIndTree<D> => {
  let res = new TNode<bigint, D>(0n, data);
  child_array.forEach((child_node: AutoIndNode<D>, ind: number) => res.children.setValue(BigInt(ind), tNode_new_k(BigInt(ind), child_node)))
  return res;
}

export function append_toplevel<K, D>(t: Tree<K, D>, n: TNode<K, D>): Tree<K, D> {
  let new_child_hm = t.children
  new_child_hm.setValue(n.id, n)

  return {
    ...t,
    children: new_child_hm,
    max_depth: t.max_depth == 0 ? 1 : t.max_depth,
    size: t.size + 1n
    // todo: update all metadata
    // like max_depth = max previous n.max_depth+1
  }
}

export function tNode_no_k<D>(d: D) {
  return new TNode<bigint, D>(0n, d)
}

export function tNode_new_k<K, D>(k: K, n: TNode<any, D>) {
  return { ...n, k: k }
}

/**
 *
 * @param t tree that shall get new child
 * @param n node to append - key gets scrubbed and auto assigned
 * @returns
 */
export function append_toplevel_autoind<D>(t: AutoIndTree<D>, n: AutoIndNode<D>): AutoIndTree<D> {
  let new_child_hm = t.children
  let new_ind = BigInt(t.children.size())
  new_child_hm.setValue(new_ind, { ...n, id: new_ind }) // TODO: test

  return {
    ...t,
    children: new_child_hm,
    // todo: update all metadata
    // (copy from above when ready)
  }
}

/**
 * TODO: fix that this function mutates t!
 * @param t
 * @returns
 */
export const append_all_toplevel_autoind = <D>(t: AutoIndTree<D>) => (ns: AutoIndNode<D>[]): AutoIndTree<D> => {
  let first_new_ind = BigInt(t.children.size())
  ns.forEach((appendee, arr_i) => {
    let new_child = tNode_new_k(first_new_ind + BigInt(arr_i), appendee)
    t.children.setValue(new_child.id, new_child)
  })

  return t
}


export const contains = <K, D>(t: Tree<K, D>) => (p: Path<K>): boolean => {
  // todo
  return false;
}
